#include "gf.h"
#include "gf_disp.h"
#include "gf_cbios.h"
#include "gf_driver.h"
#include "gf_version.h"

/*
    DEVICE_ATTR() will do PERMISSIONS Check
    Kernel Code VERIFY_OCTAL_PERMISSIONS() not allowed to Give "Other User" Write permissions
*/

#define GF_DEVICE_ATTR_RO(name) \
    static DEVICE_ATTR(name,0444,gf_##name##_show,NULL)

#define GF_DEVICE_ATTR_RW(name) \
    static DEVICE_ATTR(name,0664,gf_##name##_show,gf_##name##_store)

#define GF_DEVICE_ATTR_WO(name) \
    static DEVICE_ATTR(name,0220,NULL,gf_##name##_store)

/*
    "gf_string_show" and "gf_string_store" is demo shows
    how to add sys read write file with GF_DEVICE_ATTR_RW()
*/

static char string_buffer [64] = {0};

static ssize_t gf_string_show(struct device *dev, struct device_attribute *attr,char *buf)
{
    return sprintf(buf, "Store String:%s\n",string_buffer);
}

static ssize_t gf_string_store(struct device *dev,struct device_attribute *attr, const char *buf,size_t len)
{
    if(len>=64)
    {
        len = 63;
    }
    memcpy(string_buffer,buf,len);
    string_buffer[len] = '\0';
    return len;
}

extern int hwq_get_hwq_info(void *adp,gf_hwq_info *hwq_info);
extern int hwq_get_video_info(void *adp, gf_video_info *video_info);
static ssize_t gf_enable_usage_store(struct device *dev,struct device_attribute *attr, const char *buf,size_t len)
{
    struct pci_dev*     pdev        = container_of(dev,struct pci_dev,dev);
    struct drm_device* drm_dev      = pci_get_drvdata(pdev);
    gf_card_t*         gf_card      = drm_dev->dev_private;

    unsigned long value = 0;
    kstrtoul(buf, 0, &value);
    value = (value==0)? 0:1;
    gf_core_interface->ctl_flags_set(gf_card->adapter,1,(1UL<<16),(value<<16)); //First Int Bit16 hwq_event_enable
    if(value)
    {
        gf_core_interface->ctl_flags_set(gf_card->adapter,1,(1UL<<9),(value<<9)); //First Int Bit11, perf_event_enable
    }
    return len;
}

static ssize_t gf_enable_perf_store(struct device *dev,struct device_attribute *attr, const char *buf,size_t len)
{
    //First Int Bit15
    struct pci_dev*     pdev        = container_of(dev,struct pci_dev,dev);
    struct drm_device* drm_dev      = pci_get_drvdata(pdev);
    gf_card_t*         gf_card      = drm_dev->dev_private;

    unsigned long value = 0;
    kstrtoul(buf, 0, &value);
    value = (value==0)? 0:1;
    gf_core_interface->ctl_flags_set(gf_card->adapter,1,(1UL<<9),(value<<9)); //First Int Bit11 perf_event_enable
    if(value==0)
    {
        gf_core_interface->ctl_flags_set(gf_card->adapter,1,(1UL<<16),(value<<16)); //First Int Bit16 hwq_event_enable
    }
    return len;
}

static ssize_t gf_engine_3d_usage_show(struct device *dev, struct device_attribute *attr,char *buf)
{
    int ret = 0;
    struct pci_dev*     pdev        = container_of(dev,struct pci_dev,dev);
    struct drm_device* drm_dev      = pci_get_drvdata(pdev);
    gf_card_t*         gf_card      = drm_dev->dev_private;

    gf_hwq_info hwq_info;
    ret = hwq_get_hwq_info(gf_card->adapter,&hwq_info);

    if(ret == 0)
    {
        return sprintf(buf, "%d\n",hwq_info.Usage_3D);
    }
    return sprintf(buf, "Not Enable!\n");
}

static ssize_t gf_engine_vcp_usage_show(struct device *dev, struct device_attribute *attr,char *buf)
{
    int ret = 0;
    struct pci_dev*     pdev        = container_of(dev,struct pci_dev,dev);
    struct drm_device* drm_dev      = pci_get_drvdata(pdev);
    gf_card_t*         gf_card      = drm_dev->dev_private;

    gf_hwq_info hwq_info;
    ret = hwq_get_hwq_info(gf_card->adapter,&hwq_info);

    if(ret == 0)
    {
        return sprintf(buf, "%d\n",hwq_info.Usage_VCP);
    }
    return sprintf(buf, "Not Enable!\n");

}

static ssize_t gf_engine_vcp_info_show(struct device *dev, struct device_attribute *attr,char *buf)
{
    int ret = 0, i;
    struct pci_dev*     pdev        = container_of(dev,struct pci_dev,dev);
    struct drm_device* drm_dev      = pci_get_drvdata(pdev);
    gf_card_t*         gf_card      = drm_dev->dev_private;
    int buf_size = 0, size_offset = 0;
    int pid = 0;

    gf_sscanf(string_buffer, "%d", &pid);

    for (i = 0; i < VIDEO_INFO_NUM; i++)
    {
        gf_video_info video_info;
        gf_memset(&video_info, 0, sizeof(gf_video_info));
        video_info.index = i;

        ret = hwq_get_video_info(gf_card->adapter, &video_info);
        if (ret)
        {
            return sprintf(buf, "Not Enable!\n");
        }

        if (pid == 0) { //get all info
            if (video_info.pid[video_info.index]) {
                //PID
                size_offset = sprintf(buf + buf_size, "pid:%d\n", video_info.pid[video_info.index]);
                buf_size += size_offset;
                //clip width
                size_offset = sprintf(buf + buf_size, "width:%d\n", video_info.width[video_info.index]);
                buf_size += size_offset;
                //clip height
                size_offset = sprintf(buf + buf_size, "height:%d\n", video_info.height[video_info.index]);
                buf_size += size_offset;
                //clip codec
                size_offset = sprintf(buf + buf_size, "codec:%s\n", video_info.codec[video_info.index]);
                buf_size += size_offset;
                 //decodeapi
                size_offset = sprintf(buf + buf_size, "decode_api:%s\n", video_info.decodeapi[video_info.index]);
                buf_size += size_offset;
                //renderapi
                size_offset = sprintf(buf + buf_size, "render_api:%s\n", video_info.presentapi[video_info.index]);
                buf_size += size_offset;
                //total decode frame
                size_offset = sprintf(buf + buf_size, "total_decodeframe:%d\n", video_info.TotalDecodeFrameNum[video_info.index]);
                buf_size += size_offset;
                //total render frame
                size_offset = sprintf(buf + buf_size, "total_renderframe:%d\n", video_info.TotalRenderFrameNum[video_info.index]);
                buf_size += size_offset;
                //presentspeed
                size_offset = sprintf(buf + buf_size, "average_presentspeed:%dFPS\n", video_info.presentspeed[video_info.index]);
                buf_size += size_offset;
                //bit rate
                size_offset = sprintf(buf + buf_size, "average_bitrate:%dKbPS\n", video_info.bitrate[video_info.index]);
                buf_size += size_offset;
                //gpu deocdespeed
                size_offset = sprintf(buf + buf_size, "average_decodespeed:%dFPS\n", video_info.decodespeed[video_info.index]);
                buf_size += size_offset;
                //others
            }
        } else {  //get pid info
            if (video_info.pid[video_info.index] == pid) {
                //PID
                size_offset = sprintf(buf + buf_size, "pid:%d\n", video_info.pid[video_info.index]);
                buf_size += size_offset;
                //clip width
                size_offset = sprintf(buf + buf_size, "width:%d\n", video_info.width[video_info.index]);
                buf_size += size_offset;
                //clip height
                size_offset = sprintf(buf + buf_size, "height:%d\n", video_info.height[video_info.index]);
                buf_size += size_offset;
                //clip codec
                size_offset = sprintf(buf + buf_size, "codec:%s\n", video_info.codec[video_info.index]);
                buf_size += size_offset;
                //decodeapi
                size_offset = sprintf(buf + buf_size, "decode_api:%s\n", video_info.decodeapi[video_info.index]);
                buf_size += size_offset;
                //renderapi
                size_offset = sprintf(buf + buf_size, "render_api:%s\n", video_info.presentapi[video_info.index]);
                buf_size += size_offset;
                //total decode frame
                size_offset = sprintf(buf + buf_size, "total_decodeframe:%d\n", video_info.TotalDecodeFrameNum[video_info.index]);
                buf_size += size_offset;
                //total render frame
                size_offset = sprintf(buf + buf_size, "total_renderframe:%d\n", video_info.TotalRenderFrameNum[video_info.index]);
                buf_size += size_offset;
                //presentspeed
                size_offset = sprintf(buf + buf_size, "average_presentspeed:%dFPS\n", video_info.presentspeed[video_info.index]);
                buf_size += size_offset;
                //bit rate
                size_offset = sprintf(buf + buf_size, "average_bitrate:%dKbPS\n", video_info.bitrate[video_info.index]);
                buf_size += size_offset;
                //gpu deocdespeed
                size_offset = sprintf(buf + buf_size, "average_decodespeed:%dFPS\n", video_info.decodespeed[video_info.index]);
                buf_size += size_offset;
                //others
            }
        }

    }

    //gf_info("buffer size %d, pid %d\n", buf_size, pid);
    return buf_size;
}

static ssize_t gf_engine_vpp_usage_show(struct device *dev, struct device_attribute *attr,char *buf)
{
    int ret = 0;
    struct pci_dev*     pdev        = container_of(dev,struct pci_dev,dev);
    struct drm_device* drm_dev      = pci_get_drvdata(pdev);
    gf_card_t*         gf_card      = drm_dev->dev_private;

    gf_hwq_info hwq_info;
    ret = hwq_get_hwq_info(gf_card->adapter,&hwq_info);

    if(ret == 0)
    {
        return sprintf(buf, "%d\n",hwq_info.Usage_VPP);
    }
    return sprintf(buf, "Not Enable!\n");
}

static ssize_t gf_mclk_show(struct device *dev, struct device_attribute *attr,char *buf)
{
    struct pci_dev*     pdev        = container_of(dev,struct pci_dev,dev);
    struct drm_device* drm_dev      = pci_get_drvdata(pdev);
    gf_card_t*         gf_card      = drm_dev->dev_private;
    disp_info_t*       disp_info    = (disp_info_t*)gf_card->disp_info;

    unsigned int value = 0;
    if(S_OK == disp_cbios_get_clock(disp_info, GF_QUERY_MCLK, &value))
    {
        return sprintf(buf,"%dMHz\n", (value + 500)/1000);
    }
    return sprintf(buf, "Mclk Read Error\n");
}

static ssize_t gf_vclk_show(struct device *dev, struct device_attribute *attr,char *buf)
{
    struct pci_dev*     pdev        = container_of(dev,struct pci_dev,dev);
    struct drm_device* drm_dev      = pci_get_drvdata(pdev);
    gf_card_t*         gf_card      = drm_dev->dev_private;
    disp_info_t*       disp_info    = (disp_info_t*)gf_card->disp_info;

    unsigned int value = 0;
    if(S_OK == disp_cbios_get_clock(disp_info, GF_QUERY_VCLK, &value))
    {
        return sprintf(buf,"%dMHz\n", (value + 500)/1000);
    }

    return sprintf(buf, "Vclk Read Error\n");
}

static ssize_t gf_eclk_show(struct device *dev, struct device_attribute *attr,char *buf)
{
    struct pci_dev*     pdev        = container_of(dev,struct pci_dev,dev);
    struct drm_device* drm_dev      = pci_get_drvdata(pdev);
    gf_card_t*         gf_card      = drm_dev->dev_private;
    disp_info_t*       disp_info    = (disp_info_t*)gf_card->disp_info;

    unsigned int value = 0;
    if(S_OK == disp_cbios_get_clock(disp_info, GF_QUERY_ENGINE_CLOCK, &value))
    {
        return sprintf(buf, "%dMHz\n", (value + 500)/1000);
    }

    return sprintf(buf, "Eclk Read Error\n");

}

static ssize_t gf_free_fb_mem_show(struct device *dev, struct device_attribute *attr,char *buf)
{
    struct pci_dev*     pdev        = container_of(dev,struct pci_dev,dev);
    struct drm_device* drm_dev      = pci_get_drvdata(pdev);
    gf_card_t*         gf_card      = drm_dev->dev_private;

    gf_query_info_t  query_info;
    int mem_total = 0;
    query_info.type = GF_QUERY_SEGMENT_FREE_SIZE;
    query_info.argu = 1;
    gf_core_interface->query_info(gf_card->adapter,&query_info);
    mem_total += query_info.signed_value;

    query_info.type = GF_QUERY_SEGMENT_FREE_SIZE;
    query_info.argu = 4;
    gf_core_interface->query_info(gf_card->adapter,&query_info);
    mem_total += query_info.signed_value;

    return sprintf(buf,"%d\n",mem_total);
}

static ssize_t gf_fb_size_show(struct device *dev, struct device_attribute *attr,char *buf)
{
    struct pci_dev*     pdev        = container_of(dev,struct pci_dev,dev);
    struct drm_device* drm_dev      = pci_get_drvdata(pdev);
    gf_card_t*         gf_card      = drm_dev->dev_private;
    adapter_info_t* adapter_info    = &gf_card->adapter_info;

    return sprintf(buf,"%d M\n",adapter_info->total_mem_size_mb);
}

static ssize_t gf_slice_mask_show(struct device *dev, struct device_attribute *attr,char *buf)
{
    struct pci_dev*     pdev        = container_of(dev,struct pci_dev,dev);
    struct drm_device* drm_dev      = pci_get_drvdata(pdev);
    gf_card_t*         gf_card      = drm_dev->dev_private;
    adapter_info_t* adapter_info    = &gf_card->adapter_info;

    return sprintf(buf,"0x%x\n",adapter_info->chip_slice_mask);
}

static ssize_t gf_miu_channel_num_show(struct device *dev, struct device_attribute *attr,char *buf)
{
    struct pci_dev*     pdev        = container_of(dev,struct pci_dev,dev);
    struct drm_device* drm_dev      = pci_get_drvdata(pdev);
    gf_card_t*         gf_card      = drm_dev->dev_private;
    adapter_info_t* adapter_info    = &gf_card->adapter_info;

    return sprintf(buf,"%d\n",adapter_info->chan_num);

}

static ssize_t gf_driver_version_show(struct device *dev, struct device_attribute *attr,char *buf)
{
    return sprintf(buf,"%02x.%02x.%02x%s\n",DRIVER_MAJOR,DRIVER_MINOR,DRIVER_PATCHLEVEL,DRIVER_CLASS);
}

static ssize_t gf_release_date_show(struct device *dev, struct device_attribute *attr,char *buf)
{
    return sprintf(buf,"%s\n",DRIVER_DATE);
}

static ssize_t gf_pmp_version_show(struct device *dev, struct device_attribute *attr,char *buf)
{
    struct pci_dev*     pdev        = container_of(dev,struct pci_dev,dev);
    struct drm_device* drm_dev      = pci_get_drvdata(pdev);
    gf_card_t*         gf_card      = drm_dev->dev_private;
    disp_info_t*       disp_info    = (disp_info_t*)gf_card->disp_info;

    unsigned char* pmpversion       = disp_info->pmp_version;
    int  pmpdatelen                 = 0;
    int  pmptimelen                 = 0;

    if(*pmpversion)
    {
        // pmpVersion:str1 -> pmp version,str2 -> pmp build date,str3 -> pmp build time
        pmpdatelen = gf_strlen(pmpversion) + 1;
        pmptimelen = gf_strlen(pmpversion + pmpdatelen) + 1;
        return sprintf(buf, "%s %s %s\n",pmpversion,(pmpversion + pmpdatelen),(pmpversion  + pmpdatelen + pmptimelen) );
    }
    return sprintf(buf,"get error\n");

}

static ssize_t gf_vbios_version_show(struct device *dev, struct device_attribute *attr,char *buf)
{
    struct pci_dev*     pdev        = container_of(dev,struct pci_dev,dev);
    struct drm_device* drm_dev      = pci_get_drvdata(pdev);
    gf_card_t*         gf_card      = drm_dev->dev_private;
    disp_info_t*       disp_info    = (disp_info_t*)gf_card->disp_info;
    int  vbiosVer                   = disp_info->vbios_version;
    return sprintf(buf, "%02x.%02x.%02x.%02x\n", (vbiosVer>>24)&0xff,(vbiosVer>>16)&0xff,(vbiosVer>>8)&0xff,vbiosVer&0xff);
}

GF_DEVICE_ATTR_RW(string);
GF_DEVICE_ATTR_WO(enable_usage);
GF_DEVICE_ATTR_WO(enable_perf);

GF_DEVICE_ATTR_RO(engine_3d_usage);
GF_DEVICE_ATTR_RO(engine_vcp_usage);
GF_DEVICE_ATTR_RO(engine_vcp_info);
GF_DEVICE_ATTR_RO(engine_vpp_usage);
GF_DEVICE_ATTR_RO(vbios_version);
GF_DEVICE_ATTR_RO(pmp_version);
GF_DEVICE_ATTR_RO(release_date);
GF_DEVICE_ATTR_RO(driver_version);
GF_DEVICE_ATTR_RO(miu_channel_num);
GF_DEVICE_ATTR_RO(slice_mask);
GF_DEVICE_ATTR_RO(fb_size);
GF_DEVICE_ATTR_RO(free_fb_mem);
GF_DEVICE_ATTR_RO(eclk);
GF_DEVICE_ATTR_RO(vclk);
GF_DEVICE_ATTR_RO(mclk);

static struct attribute *gf_info_attributes[] = {
    &dev_attr_string.attr,
    &dev_attr_enable_usage.attr,
    &dev_attr_enable_perf.attr,

    &dev_attr_engine_3d_usage.attr,
    &dev_attr_engine_vpp_usage.attr,
    &dev_attr_engine_vcp_usage.attr,
    &dev_attr_engine_vcp_info.attr,
    &dev_attr_vbios_version.attr,
    &dev_attr_pmp_version.attr,
    &dev_attr_release_date.attr,
    &dev_attr_driver_version.attr,
    &dev_attr_miu_channel_num.attr,
    &dev_attr_slice_mask.attr,
    &dev_attr_fb_size.attr,
    &dev_attr_free_fb_mem.attr,
    &dev_attr_eclk.attr,
    &dev_attr_vclk.attr,
    &dev_attr_mclk.attr,
    NULL
};

const struct attribute_group gf_sysfs_group = {
    .attrs = gf_info_attributes,
    .name  = "gf_info"
};


